//
// Copyright 2015-2016, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include "BrashRedirection.h"
#include <cxxtls/stringexpressionparser.h>
#include <cxxtls/foreach.h>
#include <memory> // for auto_ptr

using namespace std;
using namespace cxxtls;

bool 
RedirType::
isNumberedRedirection(string::const_iterator &f, string::const_iterator const &l, int &fileno)
    // scan from f to l looking for on of:
    //
    //     number<
    //     number>
    //
    // If found, move f past the number, leaving it on the <>, and return the number through fileno.
    //
    // Otherwise, leave f alonge.
{
    // don't bother skipping iterators, that's done before we are to be called.

    if(f == l)
       return false;

    string::const_iterator saved_f = f;

    if(isdigit(*f))
    {
       fileno = 0;

       while(f != l && isdigit(*f))
       {
          char c = *f++;

          fileno *= 10;

          c -= '0';

          fileno += c;
       }

       if(f == l)
       {
          f = saved_f;
          return false;
       }

       if( *f == '<' || *f == '>')
       {
          return true;
       }

       f = saved_f;

       return false;

    }
    
    return false;
}

std::pair<std::string, RedirType::REDIRECTION_LIST*> 
RedirType::
parseOnlyRedirections(std::string::const_iterator &f,
                      std::string::const_iterator const &l
                     )
{
    auto_ptr<REDIRECTION_LIST> redirections(  new REDIRECTION_LIST() );

    while(f != l)
    { // parse command line tokens

        StringExpr::skipSeparators(f,l);

        while(f !=l)
        { // check for a redirections

           char c = *f;
           
           bool append = false;

           int fileNo=1;
           int otherFile=-1;  // -1 means not used

           bool numberedRedirection = RedirType::isNumberedRedirection(f, l, fileNo);

           if (f != l)
              c = *f;

           if(c == '>' || c == '<')
           {  // redirection found


              ++f;

              if(f ==l)
                 --f;
              else
              if(*f != c)
                 --f;
              else
                 append = true;

              // if we get here, we have a redirection

              ++f;  // move past c

              if(f == l)
              {
                 throw InvalidRedirection();
              }

              if(*f == '&')
              {
                 // handle "otherFile" duplications like:      2>&1
                 // which means that stderr is the same as file number stdout.

                 ++f;

                 if(f == l || !isdigit(*f))
                     throw InvalidRedirection();

                 otherFile = 0;

                 while(f != l && isdigit(*f))
                 {
                    otherFile *= 10;
                    char c = *f++;

                    otherFile += c - '0';
                 }



              }

              bool input=false;
                  
              if(c == '<')
                 input = true;


              if(otherFile != -1)
              {
                   redirections->push_back(REDIRECTION_LIST::value_type(RedirType(fileNo, input, append, otherFile), (StringExpression *)0) );
              }
              else
              {
                  StringExpr::skipSeparators(f,l);
                  
                  if(f == l)
                  {
                     throw InvalidRedirection();
                  }
                  
                  
                  auto_ptr<StringExpression> e(StringExpr::parse(f,l));
                  
                  if(!e.get())
                  {
                     throw InvalidRedirection();
                  }
                  
                  
                  
                  
                  if(!numberedRedirection)
                  {
                      if(input)
                         fileNo = 0;  
                  }
                  
                  {
                      // keep these statements together for exception safety.
                      redirections->push_back(REDIRECTION_LIST::value_type(RedirType(fileNo, input, append, otherFile), e.get()) );
                      e.release();
                  }
                  
              }

              StringExpr::skipSeparators(f,l);


           }  // redirection found
           else
              break; // no redir found

        } // check for redirection operators

        if(f == l )
           break;    // might have had a redirection at the end of the line


        StringExpr::skipSeparators(f,l);

        if(f != l)
        {
            char c = *f;

            if(c == '&' || c == ';' || c == '|' || c == '\n' || c == ')' || c == '}' )
               break;  // this an end of command terminator

            if (c == '#')
            {
                while (f != l && *f != '\n')
                   ++f;

                break;  // s'all good if we end with a comment
            }
            
        }

        string errorMessage="Unexpected token in redirection list: ";

        errorMessage += *f;

        return pair<string,REDIRECTION_LIST*>(errorMessage,reinterpret_cast<REDIRECTION_LIST*>(0));


    } // parse command line tokens


    return pair<string,REDIRECTION_LIST*>(string(),redirections.release());

}


void 
RedirType::
redirectFileHandles(cxxtls::ScriptThread::ThreadContext *workingContext_, RedirType::REDIRECTION_LIST const *redirs )
{
        typedef RedirType::REDIRECTION_LIST::value_type redirPair;
         
        CXXTLS_FOREACH(redirPair const &cur, *redirs)
        {
            StringArray output;

            if(cur.second)
                cur.second->execute(&workingContext_->variables(), output);

            string redirectedFileName = cur.second ? output.combine() : string();

            if(!redirectedFileName.empty())
            {
                // handle ~ expansions -- noting that the file name can contain 1>~/log.txt


               size_t offset = redirectedFileName.find_first_not_of("0123456789<>");

               if(offset < redirectedFileName.size()-1)  // size guaranteed not to be zero by if statement, above
               {
                   if(redirectedFileName[offset] == '~')
                   {
                       char c = redirectedFileName[offset+1];

                       if(c == '/' || c == '\\')
                       {
                           string homedir = workingContext_->variables().find("BRASH_USER_HOME");

                           if(!homedir.empty())
                           {
                               redirectedFileName = redirectedFileName.substr(0,offset) + homedir + redirectedFileName.substr(offset+1);
                           }
                       }
                   }
               }

            }


        
            int fileNo    = cur.first.fileNo_;
            int otherFile = cur.first.otherFile_;
            bool isInput  = cur.first.isInput_;
            bool isAppend = cur.first.isAppend_;

            // struct NonAllowedRedirectionException{};      // cannot swap input and output
            // struct EmptyRedirectionFilenameException{};   // some operations must have a name
            // struct CannotDupBadFileException{};           // attempt to duplicate a bad file handle
            // struct CannotOpenTargetFileException{};       // just can't open the desired file

           ScriptThread::ManagedFileHandles::FileHandleAttributes   extantAttributes = workingContext_->fileHandles_->getFileAttributes(fileNo);

           if(extantAttributes.good())
           {
              if(extantAttributes.input_ != isInput)
                  throw NonAllowedRedirectionException();
           }

           if(otherFile >= 0)
           {
               // dup operation, not opening a named file

               if(!workingContext_->fileHandles_->reopenDup(fileNo, otherFile))
               {
                   throw CannotDupBadFileException();
               }

           }
           else
           {
              // normal file reopen operation.

              // redirected name may or may not contain things like <, 14> 2>> at the begining
              // so, we are going to skip passed everything up to the first >> or << group
              // and remove them

              size_t bracketOffset = redirectedFileName.find_first_of("<>");

              if(bracketOffset < redirectedFileName.size() )
              {
                  if(isAppend)
                     ++bracketOffset;

                  ++bracketOffset;  // now points to the first character in the actual file name

                  redirectedFileName.erase(0,bracketOffset);
              }


              if (!(isAppend && isInput))
              {
                  // don't come in here if we are <<*eof or some such thing
                  // we just want to complete ignore <<*eof style redirections as they 
                  // must be  handled entirely by the parser -- basically, this kind of redirection
                  // is actually a piping operation under the covers.
              
                  if(!workingContext_->fileHandles_->reopenAs(fileNo, redirectedFileName, isAppend))
                  {
                      throw CannotOpenTargetFileException("filename", redirectedFileName);
                  }
              }

           }
        
        }
}
