//
// Copyright 2015 Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

#include <iostream>
#include <map>
#include <string>
#include <vector>
#include <exception>

namespace cxxtls
{

namespace CommandLine
{

struct OptionException
:  public std::exception
   ///  An exception thrown by the constructor to the Options class below.  You must catch
   ///  this exception to handle user syntax errors.  The text of the what() message is
   ///  fully explanator of the mistake the user has made.
{
    std::string optionName_;  

    virtual char const *what()  const  throw()
    {
        return optionName_.c_str();
    }

    OptionException(char const *name, char const *cause=0)
    {
       optionName_ = std::string("Invalid command line option :  '-") + name + std::string("'");

       if(cause)
       {
          optionName_ += " (";
          optionName_ += cause;
          optionName_ += ")";
       }
    }

    ~OptionException() throw()
    {
    }
};

struct OptionDeclaration
    ///  An object that describes an allowed command line option.  
    ///
    ///  The constructor takes the following parameters:
    ///
    ///   @li name -- the name of the option without leading dashes, e.g.  "o", or "file"
    ///   @li parms -- the count of allowed data values to follow the parameter.
    ///                <ul>
    ///                   @li 0 -- means no values are allowed.  This is a strictly boolean option.
    ///                   @li 1 -- means exactly 1 value is allowed, and it is required.
    ///                   @li 2 -- means exactly 1 value is allowed, but the option is allowed 
    ///                            to be used repeatedly to enable the build up of a list of values
    ///                            for the parameter.
    ///                </ul>
    ///   @li mode --  Indicates which style of option usage is required for this option.  
    ///                The styles are as follows:
    ///@begincode
    ///
    ///    mode 0:   
    ///               Option names must be a single character.  A single dash must be used on the 
    ///               command line to identify the argument as an option.  Options which do not 
    ///               accept parameters may be grouped, but options with parameters  must be used 
    ///               singly on the command line.  Data parameters to the option can occur either 
    ///               as part of the option declaration or as a separate command line parameter 
    ///               string.  Here are some example mode 0 uses:
    ///
    ///                   -opqr   # valid only if none of o,p,q, or r have a non-zero parm count.
    ///
    ///                   -I/usr/include   # valid if -I has a parms value which is non-zero
    ///
    ///                   -I  /usr/include # same as above
    ///
    ///    mode 1:    Option names can be of any length.  Options may not becombined into a single 
    ///               parameter.  Options must begin with 2 dashes on the command line.  Optional 
    ///               parameter data may be specifed in either of two ways:  either as a separator 
    ///               parameter, or combined with the option name using the equal character as a 
    ///               separator.  Here are some example usages:
    ///
    ///                   --OptionName
    ///                   --OptName=value
    ///                   --Opt=value1  --Opt=value2 -opt=value3   # only works if parms == 2 in the 
    ///                                                            # OptionDeclaration
    ///                
    ///@endcode
{
   std::string name_;    ///< Example:  "o" -- as in "-o" or "--o"
   char        parms_;   ///< Example:  0, 1, or 2.  Where   0 means no parameters, 1 means 1 parameter, 2 means many parameters
   char        mode_;    ///< Example:  0, 1         Where   0 means only a character name, specifiable in groups like -abc. 1 means --style
   std::string help_;

   OptionDeclaration(std::string const &name, char parms, char mode, std::string const &help)
   :  name_(name)
   ,  parms_(parms)
   ,  mode_(mode)
   ,  help_(help)
       ///   Construct an OptionDeclaration object.  See the class comments for more details. 
   {
   }
};

struct Options
    ///  A utility class that lets you parse a program's command line arguments and keep them in a data
    ///  structure.
    ///
    ///  A list of allowed options must be specified as an option to the constructor.  See class 
    ///  OptionDeclaration for more details on how options are used.
    ///
    ///  The constructor copies the command line arguments into vector and splits apart the "flags"
    ///  from the positional parameters.  The positional parameters can be obtained by calling the
    ///  arg() method given a numberical index between 0 and the last argument number.  To get a count
    ///  of the position parameters, call the size() method.
    ///
    ///  To see if flag has been set, use isSet().  To see if the value of the flag, if one is specified,
    ///  use flagValue().  Note that flagValue() returns a reference to a vector of string, not a single
    ///  string.  The vector may be empty or may have more than one value in it, based on the 
    ///  OptionDeclaration object for that option name.
    ///
{
    typedef std::map< std::string, std::vector<std::string > >  FlagMap;

private:

    FlagMap                         flags_;
    std::vector<std::string>        arguments_;
    std::vector<OptionDeclaration>  optionInfo_;

    static std::vector<std::string> emptyVector_;

public:

    bool isSet(std::string const &name) const 
        ///  @return true if the specified flag is set
        ///
        ///  @note Do not include leading dashes in the name!
    { 
       return flags_.find(name) != flags_.end(); 
    }

    size_t size() const 
        ///   @return the count of positional parameters (including argument 0 which is the program name)
    { 
        return arguments_.size(); 
    }

    size_t empty() const 
        ///   @return true if there are no normal command line arguments
    { 
        return arguments_.empty(); 
    }


    std::string arg(size_t i) const
        ///  @return the value of a positional parameter.  use index 0 for the program name.
    { 
        if(i < size()) 
            return arguments_[i]; 
        else 
            return std::string(); 
    }

    std::vector<OptionDeclaration> const &info() const 
        ///  @return a reference to the list of options allowed
    {
       return optionInfo_;
    }

    OptionDeclaration const *optionDeclaration(std::string const &name) const;
        ///< @return null if the name is not a valid option or
        /// return a pointer to the OptionDeclaration object if it is.

    std::vector<std::string> const &flagValue(std::string const &flagName) const
        ///   @return a reference the vector holding the values which have been
        ///    added for a given flag name by all uses of it on the command line.
        ///
    {
        FlagMap::const_iterator where = flags_.find(flagName);

        if(where == flags_.end())
           return emptyVector_;

        return where->second;
    }

    Options(char **argv, OptionDeclaration *info, size_t count)
        ///   Construct an Options argument by parsing the command line vector in
        ///   context of the list of allowed options.
        ///   Throw an exception if the user specifies an invalid option on the command line.
        ///
        ///  @param args   The pointer to the program's command line arguments -- just like you get from
        ///                from the invocation of main.  Note that the arguments must be terminated with a
        ///                null pointer.
        ///
        ///  @param info   A pointer to the first member of an array of OptionDeclaration objects which
        ///                defines the valid options and their attributes.
        ///
        ///  @param count  The count of the OptiondDeclaration objects refered to by parameter info.
    {
        std::map<std::string,bool>  nameset;   // could be a set but I'll use a map to avoid #including <set>


        while(count--)
        {
           if(nameset.find(info->name_) != nameset.end())
           {
               throw OptionException("duplicate option", info->name_.c_str());
           }

           nameset.insert(std::pair<std::string,bool>(info->name_, true));

           optionInfo_.push_back(*info++); 
        }

        std::vector<std::string> tmp;

        while(*argv)
        {
            tmp.push_back(*argv++);
        }

        populate(tmp.begin(), tmp.end());
    }

    template<class Iterator>
    Options(Iterator f, Iterator l, OptionDeclaration *info, size_t count)
        ///   Construct an Options argument by parsing the command line vector in
        ///   context of the list of allowed options.
        ///   Throw an exception if the user specifies an invalid option on the command line.
        ///
        ///  @param f      An iterator to the first string in the options (should point to program name)
        ///
        ///  @param l      An iterator to the last string (points past the last valid string)
        ///
        ///  @param info   A pointer to the first member of an array of OptionDeclaration objects which
        ///                defines the valid options and their attributes.
        ///
        ///  @param count  The count of the OptiondDeclaration objects refered to by parameter info.
    {
        while(count--)
        {
           optionInfo_.push_back(*info++); 
        }

        populate(f,l);
    }


    std::ostream &printHelp(std::ostream &s) const;
        ///<  Print a description of the options to the specified stream.
        ///   This function is meant as a component of a program's
        ///   help message display.



    template<class Iterator>
    void populate(Iterator f, Iterator l)
        ///  parse the command line arguments and store them in *this.
    {
        // handle flags

        if(f != l)
            arguments_.push_back(*f++);   // push the program name onto the arguments list


        // read and handle the rest of the arguments on the command line.

        // basically, while there are leading arguments whose first character is '-', treat
        // them as flags.  When we hit an argument that does not begin with '-', then we switch
        // a mode of just copying the arguments into the arguments_ vector.
        //
        // an exception to the above occurs when an argument has a parameter that is not combined
        // into the argument itself.  Here's a couple of examples:
        //
        //   -t parm -u -v arg1 arg2 arg3
        //
        //   --File name --Single arg1 arg2 arg3
        //
        //  Arguments "parm", and "name" do not terminate the parsing of flags because they are
        //  consumed by the interpretation of "-t" and "--file".

        while(f != l)
        {
            std::string cur = *f++;

            if(cur.empty())
            {
               --f;
               break;
            }

            char c = cur[0];

            if(c == '-')
            {
                if(cur.size() == 1)
                    break;  // a single - is a the end of the flags
                else
                if(cur.size() == 2 && cur[1] == '-')
                    break;  // a -- by itself is also the end of the flags
                else
                    handleFlag(cur, f, l); 
            }
            else
            {
               --f;
               break;  // this is the first non-flag argument
            }

        }

        // handle the rest of the arguments

        while(f != l)
        {
            arguments_.push_back(*f++);
        }
    }

    template<class Iterator>
    void handleFlag(std::string const &cur, Iterator &f, Iterator &l)
       /// handle the interpretation of each flag style option -- including the possibility
       /// of it having an option.  Throw an error if the flag's name is invalid or if
       /// the specification is not valid give the OptionDeclaration passed to the constructor.
    {

        std::string::const_iterator a = cur.begin(), b = cur.end();

         // count the dashes used on the command line

        int dashCount=0;

        while(a != b)
        {
            if(*a != '-')
               break;

            ++dashCount;
            ++a;
           
        }

        // validate the dash count

        if(dashCount < 1 || dashCount > 2)
        {
           throw OptionException(cur.c_str(), "wrong number of dashes used in option reference");
        }


        // handle options defined by single dashes, and below, those by 2 dashes.

        if(dashCount == 1)
        {
            // we are dealing with 1 char option names: a, b, c, etc.

            bool nonDataOptionFound = false;
            bool dataOptionFound    = false;


            while(a !=b)
            {
                // iterate over the characters in cur, one at a time, treating each as an option
                // name and making all needed validataions.  When data is found, attach it to the
                // option of interest.

                char buffer[2];       // treat the character as a string

                buffer[0] = *a++;
                buffer[1] = 0;

                OptionDeclaration const *info = optionDeclaration(buffer);  // confirm that it is a valid option name

                if(info == 0)
                {
                    throw OptionException(buffer);
                }

                // validate option as proper member of the current group.

                if(info->parms_ == 0)
                {
                   nonDataOptionFound = true;
                }
                else
                {
                   dataOptionFound = true;
                }

                if(dataOptionFound && nonDataOptionFound)
                {
                   throw OptionException(buffer, "cannot mix options with data and options without it");
                }


                if(info->mode_)
                {
                   // bummer dude, like, like, like man this option is only supposed to be used like this:  --o
                   //
                   // but the user did it like this:  -o

                   throw OptionException(buffer, "wrong number of dashes used for option");
                }

                if (!isSet(buffer))
                    flags_[buffer]=emptyVector_;      // currently ignoring option parameters, but it should be here

                if(info->parms_)
                {
                   // there should be exactly 1 parameter.  If parms_ is > 1, it means that multiple -option value specs are allowed, not that multiple
                   // data values are supplied.

                   if(f == l && a == b)
                   {
                       throw OptionException(buffer, "expected parameter");
                   }


                   if(info->parms_ == 1)
                   {
                      if(!flags_[buffer].empty())
                         throw OptionException(buffer, "too many uses of the option");
                   }

                   if(a != b)
                   {
                       std::string tmp;

                       while(a != b)
                          tmp += *a++;

                       flags_[buffer].push_back(tmp);
                   }
                   else
                       flags_[buffer].push_back(*f++);

                }

            }
        }
        else
        {
           // we are dealing with options that look like this:   --fred=bill

           std::string name;

           while(a != b && *a != '=')
           {
               name += *a++;
           }

           OptionDeclaration const *info = optionDeclaration(name);

           if(!info)
              throw OptionException(name.c_str(), "no such option");

           if (info->mode_ == 0)
               throw OptionException(name.c_str(), "wrong number of dashes used on option");

           if(!isSet(name))
              flags_[name]=emptyVector_;

           if(a != b)
           {
              if(info->parms_  == 0)
                 throw OptionException(name.c_str(), "no parameter allowed");  // you are not supposed to have parameters
              else
              if(info->parms_ == 1 && !flags_[name].empty())
                 throw OptionException(name.c_str(), "too many uses of the option");  // you are only supposed to have 1 parameter for all usages

              // multiple parm settings are ok (in multiple parm invocations, not in the same invocation)

              std::string value;

              ++a;  // skip =

              while(a != b)
                 value += *a++;

             flags_[name].push_back(value);

             // note:  here we are not incrementing f because the option already has the = sign in it
             //  although theoretically, we could support both modes -- and say gobble the value out of the
             //  command line... hmmmmmmm....

           }
           else
		   if (info->parms_)
           {
               // we have something like this:   --sep ,

               if (f == l)
               {
                   flags_[name].push_back("");

               }
               else
                   flags_[name].push_back(*f++); 
           }
       }

    }


    
}; // struct Options

}  // namespace CommandLine

}  // namespace cxxtls
