#ifndef cxxtls_mru_h_included
#define cxxtls_mru_h_included
//
// Copyright 2012, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.

#include <list>

namespace cxxtls
{
    template<class T>
    class MRU
        ///  A template container designed to provide a lightweight "most recently used"
        ///  functionality.  "Least recently used" is also implemented.
        ///
        ///  Key features:
        ///
        ///  @li use activate(T) to both add items and promote them to be the
        ///                      most recently used.
        ///
        ///  @li use discard(T)  to remove items from the container
        ///
        ///  @li use mostRecent() to get an iterator to the most recently activated
        ///                       item.  May return end..
        ///
        ///  @li use leastRecent() to get an iterator to the least recently 
        ///                        activated item.  May return end().
        ///
        ///  @li empty() tells you if the container is empty, as does the 
        ///              expression @b mostRecent() @b == @b end().
        ///
        ///  The general expectation is that applications will use this container to
        ///  access no more than a few hundred items.  A hash_list could be used for
        ///  the rep_type if more than this will be needed.  It provides O(1) amortized
        ///  lookups which will speed up finds.
    {
    public:

        typedef            T   value_type;
        typedef std::list<T>   rep_type;

        typedef typename rep_type::iterator        iterator;

        iterator end()          { return data_.end();   }  ///< @return the end of the container

        iterator mostRecent()   { return data_.begin(); }  ///< @return most recently used or end()

        iterator leastRecent()  
            ///  @return the least recently used item or end()
        { 
            if(empty()) 
                return data_.end();
            
            iterator rv = data_.end();
            --rv;

            return rv;
        }

        bool empty() const { return data_.empty(); }  ///< @return true if container is empty

        void activate(T const &item)
           /// Find the item of interest and mark it as the most recently used
           /// or add it to the container as the most recently used.
        {
            iterator where = find(item);

            if(where != data_.end())
            {
               data_.erase(where);
            }

            data_.push_front(item);

        }

        void discard(T const &item)
            ///  Remove the specified item from the container
        {
            iterator where = find(item);

            if(where != data_.end())
               data_.erase(where);
        }



    private:
        rep_type data_;

        iterator find(T const &query) 
            //  return an iterator pointing to query or end() if is not found
        {
           iterator next = data_.begin();
           iterator end  = data_.end();

           while(next != end)
           {
               if(query == *next)
                   return next;
                   
               ++next;
           }

           return next;

        }



    };  // class MRU

}; // namespace cxxtls


#endif
